@@ -2,12 +2,11 @@ from __future__ import annotations
 
 from typing import TYPE_CHECKING
 from typing import Protocol
-from typing import Sequence
 
 if TYPE_CHECKING:
     from sqlframe.base.column import Column
 
+    from narwhals._spark_like.utils import WindowInputs
+
     class WindowFunction(Protocol):
-        def __call__(
-            self, _input: Column, partition_by: Sequence[str], order_by: Sequence[str]
-        ) -> Column: ...
+        def __call__(self, window_inputs: WindowInputs, /) -> Column: ...
